1 VERSIONINFO
FILEVERSION    	$VERSION_CSV
PRODUCTVERSION 	$VERSION_CSV
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4"
        BEGIN
            VALUE "CompanyName",      "Google Inc."
            VALUE "FileDescription",  "PDFium"
            VALUE "FileVersion",      "$VERSION"
            VALUE "InternalName",     "pdfium"
            VALUE "OriginalFilename", "pdfium.dll"
            VALUE "ProductName",      "pdfium"
            VALUE "ProductVersion",   "$VERSION"
            VALUE "LegalCopyright",   "Copyright 2025-present PDFium Authors."
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1252
    END
END
