# SPDX-FileCopyrightText: 2025 geisserml <geisserml@gmail.com>
# SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause

# platform files are managed in an own pypdfium2_raw module, cleanly separating raw bindings and helpers
from pypdfium2_raw.bindings import *  # noqa
